/*
NAME:    centerOfMass

PURPOSE: This program calculates the center of mass for a system of
particles. Each particle P_i with mass m_i is located in a space
with coordinates r_i for i=1,..,n. The coordinate of center of
mass is R and it can be obtained by solving this equation:

  R = [(m_1 * r_1) + ....  + (m_n * r_n)] / M

where, M = Sum of masses of all particles

USAGE:   The program takes as input the number of particles

  program <# of threads> <# of particles>

The output consists of diagnostics to make sure the algorithm worked,
and of timing statistics.

FUNCTIONS CALLED:

Other than OpenMP or standard C functions, the following
functions are used in this program:


HISTORY: KP (HPCL GWU)
 ***********************************************************************************/

#include<omp.h>
#include<stdio.h>
#include<stdlib.h>
#include<assert.h>

typedef struct particles_s
{
  //coordinates
  double *x;
  double *y;
  double *z;
  //mass
  double *m;
} particles_t;


int main(int argc,char *argv[])
{

  int i;

  if ( argc != 3)
  {
    printf("Usage: program <# of threads> <# of particles> \n");
    exit(1);
  }

  // print arguments
  printf("Default Number of threads: %d \n",omp_get_num_threads());

  long int threads = atoi(argv[1]);
  omp_set_num_threads(threads);
  printf("Number of threads: %lu \n",threads);

  unsigned long int N = strtoul((argv[2]),NULL, 10);
  printf("Number of particles: %lu \n",N);

  printf("-----------------------------\n");

  particles_t particle;
  particle.x = malloc(sizeof(double)*N);
  particle.y = malloc(sizeof(double)*N);
  particle.z = malloc(sizeof(double)*N);
  particle.m = malloc(sizeof(double)*N);
  assert(particle.x &&
         particle.y &&
         particle.z &&
         particle.m);

  // intialize center of mass
  double com_x=0,com_y=0,com_z=0,total_mass=0.1;
  double start_wtime, end_wtime;

  // Initialize rand number generator
  srand(21);


  // fork into multiple threads.
  // each thread has its private copy of i
#pragma omp parallel shared(particle) private(i)
  {

    // assign Random coordinates and mass to each particle
#pragma omp for  private(i)
    for(i=0;i<N;i++)
    {
      particle.x[i] = rand() % 100;
      particle.y[i] = rand() % 100;
      particle.z[i] = rand() % 100;
      particle.m[i] = rand() % 100;
    }

#pragma omp master
    {
      // start clock
      start_wtime = omp_get_wtime();
    }
#pragma omp barrier
#pragma omp for  private(i) reduction(+:total_mass) nowait
    for(i=0;i<N;i++)
    {
      total_mass += particle.m[i];
    }

    // calculate center of mass for N particles
#pragma omp for  private(i) reduction(+:com_x,com_y,com_z) nowait
    for(i=0;i<N;i++)
    {
      com_x += particle.x[i] * particle.m[i]/total_mass;
      com_y += particle.y[i] * particle.m[i]/total_mass;
      com_z += particle.z[i] * particle.m[i]/total_mass;
    }
  }

  // calculate total time
  end_wtime = omp_get_wtime();
  double total_time = (end_wtime - start_wtime );

  printf(" Time taken %lf seconds \n" ,total_time);
  printf ("The center of mass is at (%lf ,%lf) \n", com_x,com_y);
  return 0;
}
